
/* 

	-------------------------------------------------------------------------------------------------------------------------------------------------------
	-------------------------------------------------------------------------------------------------------------------------------------------------------

						Stefano Mattoccia

						DISI - University of Bologna

						www.vision.deis.unibo.it/smatt

						stefano.mattoccia@unibo.it

						Summer 2010

						If you use this code please cite the following paper:

						S. Mattoccia, S. Giardino, A. Gambini, 
						"Accurate and efficient cost aggregation strategy for stereo correspondence based on approximated joint bilateral filtering"
						Asian Conference on Computer Vision (ACCV2009), September 23- 27, 2009,  Xi'an, China


	-------------------------------------------------------------------------------------------------------------------------------------------------------
	-------------------------------------------------------------------------------------------------------------------------------------------------------

*/


#define SMP_ON


#include "ACCV2009.h"



float EXP_LUT_ACCV_2009[3*256*256];

int TAD(int value, int threshold)
{

 if (value<threshold) 
  return value;
 else 
  return threshold; 

}



double get_subpixel_adjustment_min(double left, double center, double right)
{
	// Subpixel interpolation

	double adjustment=0;

	if ((right+left-2.0*center)!=0)
		adjustment=-((double)right - (double)left)/(2.0*((double)right+(double)left-(double)2.0*center));

	return adjustment;
}



// -----------------------------------------------------------------------------------------------------------------------
// -----------------------------------------------------------------------------------------------------------------------
// -----------------------------------------------------------------------------------------------------------------------


int init_exponential_lut_ACCV_2009(float gamma_color, int Dim)
{
	for (int i=0; i<Dim; i++)
		EXP_LUT_ACCV_2009[i]=exp(-(sqrt((float)i)/gamma_color));

	return 0;
}


// -----------------------------------------------------------------------------------------------------------------------
// -----------------------------------------------------------------------------------------------------------------------
// -----------------------------------------------------------------------------------------------------------------------



void box_filtering_stereo_color_TAD_single_channel_ACCV_2009(int w, int h, int nDisp, int offs_x, IplImage *L, IplImage *R, int** out, int r, int threshold)
{
	int x,y,i,j,d;
	int *acc = (int *) calloc(nDisp, sizeof(int));
	int *row1 = (int *) calloc(nDisp, sizeof(int));
	int *row2 = (int *) calloc(nDisp, sizeof(int));
	int *pos0 = (int *) calloc(nDisp, sizeof(int));

	int **V;
	V = (int **)calloc(w, sizeof(int*));

	for(d=0; d<w; d++){
		V[d] = (int *)calloc(nDisp, sizeof(int)); 
	}

	int n = 2*r+1;
	int sqn = n*n;

	//############### PRIMA RIGA #################
	//FASE "0": PRIMA POSIZIONE A SX
	for(d=0; d<nDisp; d++){
		V[0][d]=0;
		for(j=0; j<r+1; j++)
			V[0][d] += 
			TAD(abs(((unsigned char)L->imageData[j*L->widthStep])-((unsigned char)R->imageData[j*R->widthStep])) 
			+ abs(((unsigned char)L->imageData[j*L->widthStep+1])-((unsigned char)R->imageData[j*R->widthStep+1])) + 
			abs(((unsigned char)L->imageData[j*L->widthStep+2])-((unsigned char)R->imageData[j*R->widthStep+2])),threshold);

		acc[d]=V[0][d];

		for(i=1; i<r+1; i++)
			for(j=0; j<r+1; j++)
				acc[d] += 
				TAD(abs(((unsigned char)L->imageData[j*L->widthStep+3*i])-((unsigned char)R->imageData[j*R->widthStep+3*(i-min(i,d+offs_x))])) 
				+ abs(((unsigned char)L->imageData[j*L->widthStep+3*i+1])-((unsigned char)R->imageData[j*R->widthStep+3*(i-min(i,d+offs_x))+1])) 
				+ abs(((unsigned char)L->imageData[j*L->widthStep+3*i+2])-((unsigned char)R->imageData[j*R->widthStep+3*(i-min(i,d+offs_x))+2])),threshold);
		out[0][d] = acc[d];

		pos0[d] = acc[d];
	}

	//FASE 1: bordo sinistro
	for(x=1; x<r+1; x++){
		for(d=0; d<nDisp; d++){
			V[x+r][d]=0;
			for(j=0; j<r+1; j++)
				V[x+r][d] += 
				TAD(abs(((unsigned char)L->imageData[j*L->widthStep+3*(x+r)])-((unsigned char)R->imageData[j*R->widthStep+3*(x+r-min(x+r,d+offs_x))]))
				+ abs(((unsigned char)L->imageData[j*L->widthStep+3*(x+r)+1])-((unsigned char)R->imageData[j*R->widthStep+3*(x+r-min(x+r,d+offs_x))+1])) 
				+ abs(((unsigned char)L->imageData[j*L->widthStep+3*(x+r)+2])-((unsigned char)R->imageData[j*R->widthStep+3*(x+r-min(x+r,d+offs_x))+2])),threshold);

			acc[d] = acc[d] + V[x+r][d];
			out[x][d] = acc[d];
			V[x][d]=0;
			for(j=0; j<r+1; j++)
				V[x][d] += 
				TAD(abs(((unsigned char)L->imageData[j*L->widthStep+3*x])-((unsigned char)R->imageData[j*R->widthStep+3*(x-min(x,d+offs_x))])) 
				+ abs(((unsigned char)L->imageData[j*L->widthStep+3*x+1])-((unsigned char)R->imageData[j*R->widthStep+3*(x-min(x,d+offs_x))+1])) 
				+ abs(((unsigned char)L->imageData[j*L->widthStep+3*x+2])-((unsigned char)R->imageData[j*R->widthStep+3*(x-min(x,d+offs_x))+2])),threshold);
		}//d
	}//x

	//FASE 2: parte centrale
	for(x=r+1; x<w-r; x++){
		for(d=0; d<nDisp; d++){

			V[x+r][d]=0;
			for(j=0; j<r+1; j++)
				V[x+r][d] += 
				TAD(abs(((unsigned char)L->imageData[j*L->widthStep+3*(x+r)])-((unsigned char)R->imageData[j*R->widthStep+3*(x+r-min(x+r,d+offs_x))])) 
				+ abs(((unsigned char)L->imageData[j*L->widthStep+3*(x+r)+1])-((unsigned char)R->imageData[j*R->widthStep+3*(x+r-min(x+r,d+offs_x))+1])) 
				+ abs(((unsigned char)L->imageData[j*L->widthStep+3*(x+r)+2])-((unsigned char)R->imageData[j*R->widthStep+3*(x+r-min(x+r,d+offs_x))+2])),threshold);

			acc[d] = acc[d] + V[x+r][d] - V[x-r-1][d];
			out[x][d] = acc[d];
		}//d
	}//x

	//FASE 3: bordo destro
	for(x=w-r; x<w; x++){
		for(d=0; d<nDisp; d++){

			acc[d] = acc[d] - V[x-r-1][d];
			out[x][d] = acc[d];
		}//d
	}//x

	//######## BORDO IN ALTO ############
	for(y=1; y<r+1; y++){
		//FASE "0": PRIMA POSIZIONE A SX
		for(d=0; d<nDisp; d++){
			V[0][d] = V[0][d] + 
				TAD(abs(((unsigned char)L->imageData[(y+r)*L->widthStep])-((unsigned char)R->imageData[(y+r)*R->widthStep])) 
				+ abs(((unsigned char)L->imageData[(y+r)*L->widthStep+1])-((unsigned char)R->imageData[(y+r)*R->widthStep+1])) 
				+ abs(((unsigned char)L->imageData[(y+r)*L->widthStep+2])-((unsigned char)R->imageData[(y+r)*R->widthStep+2])),threshold);

			row2[d] = 0;
			for(i=0; i<r+1; i++)
				row2[d] += 
				TAD(abs(((unsigned char)L->imageData[(y+r)*L->widthStep+3*i])-((unsigned char)R->imageData[(y+r)*R->widthStep+3*(i-min(i,d+offs_x))])) 
				+ abs(((unsigned char)L->imageData[(y+r)*L->widthStep+3*i+1])-((unsigned char)R->imageData[(y+r)*R->widthStep+3*(i-min(i,d+offs_x))+1])) 
				+ abs(((unsigned char)L->imageData[(y+r)*L->widthStep+3*i+2])-((unsigned char)R->imageData[(y+r)*R->widthStep+3*(i-min(i,d+offs_x))+2])),threshold);

			acc[d] = pos0[d] + row2[d];
			pos0[d] = acc[d];

			out[y*w][d] = acc[d];
		}

		//FASE 1: bordo sinistro 
		for(x=1; x<r+1; x++){
			for(d=0; d<nDisp; d++){
				V[x+r][d] = V[x+r][d] + 
					TAD(abs(((unsigned char)L->imageData[(y+r)*L->widthStep+3*(x+r)])-((unsigned char)R->imageData[(y+r)*R->widthStep+3*(x+r-min(x+r,d+offs_x))])) 
					+ abs(((unsigned char)L->imageData[(y+r)*L->widthStep+3*(x+r)+1])-((unsigned char)R->imageData[(y+r)*R->widthStep+3*(x+r-min(x+r,d+offs_x))+1])) 
					+ abs(((unsigned char)L->imageData[(y+r)*L->widthStep+3*(x+r)+2])-((unsigned char)R->imageData[(y+r)*R->widthStep+3*(x+r-min(x+r,d+offs_x))+2])),threshold);

				acc[d] = acc[d] + V[x+r][d];
				out[y*w+x][d] = acc[d];
				V[x][d] = V[x][d] + 
					TAD(abs(((unsigned char)L->imageData[(y+r)*L->widthStep+3*x])-((unsigned char)R->imageData[(y+r)*R->widthStep+3*(x-min(x,d+offs_x))])) + 
					abs(((unsigned char)L->imageData[(y+r)*L->widthStep+3*x+1])-((unsigned char)R->imageData[(y+r)*R->widthStep+3*(x-min(x,d+offs_x))+1])) + 
					abs(((unsigned char)L->imageData[(y+r)*L->widthStep+3*x+2])-((unsigned char)R->imageData[(y+r)*R->widthStep+3*(x-min(x,d+offs_x))+2])),threshold);
			}
		}//x

		//FASE 2: parte centrale
		for(x=r+1; x<w-r; x++){
			for(d=0; d<nDisp; d++){
				V[x+r][d] = V[x+r][d] + 
					TAD(abs(((unsigned char)L->imageData[(y+r)*L->widthStep+3*(x+r)])-((unsigned char)R->imageData[(y+r)*R->widthStep+3*(x+r-min(x+r,d+offs_x))])) + 
					abs(((unsigned char)L->imageData[(y+r)*L->widthStep+3*(x+r)+1])-((unsigned char)R->imageData[(y+r)*R->widthStep+3*(x+r-min(x+r,d+offs_x))+1])) + 
					abs(((unsigned char)L->imageData[(y+r)*L->widthStep+3*(x+r)+2])-((unsigned char)R->imageData[(y+r)*R->widthStep+3*(x+r-min(x+r,d+offs_x))+2])),threshold);

				acc[d] = acc[d] + V[x+r][d] - V[x-r-1][d];
				out[y*w+x][d] = acc[d];
			}
		}//x

		//FASE 3: bordo destro
		for(x=w-r; x<w; x++){
			for(d=0; d<nDisp; d++){
				acc[d] = acc[d] - V[x-r-1][d];
				out[y*w+x][d] = acc[d];
			}
		}//x
	}//y

	//######## PARTE CENTRALE ############
	for(y=r+1; y<h-r; y++){

		//FASE "0": PRIMA POSIZIONE A SX
		for(d=0; d<nDisp; d++){
			V[0][d] = V[0][d] 
			- ( TAD(abs(((unsigned char)L->imageData[(y-r-1)*L->widthStep])-((unsigned char)R->imageData[(y-r-1)*R->widthStep])) + 
				abs(((unsigned char)L->imageData[(y-r-1)*L->widthStep+1])-((unsigned char)R->imageData[(y-r-1)*R->widthStep+1])) + 
				abs(((unsigned char)L->imageData[(y-r-1)*L->widthStep+2])-((unsigned char)R->imageData[(y-r-1)*R->widthStep+2])),threshold)) 
				+ 
				TAD(abs(((unsigned char)L->imageData[(y+r)*L->widthStep])-((unsigned char)R->imageData[(y+r)*R->widthStep])) + 
				abs(((unsigned char)L->imageData[(y+r)*L->widthStep+1])-((unsigned char)R->imageData[(y+r)*R->widthStep+1])) + 
				abs(((unsigned char)L->imageData[(y+r)*L->widthStep+2])-((unsigned char)R->imageData[(y+r)*R->widthStep+2])),threshold);


			row1[d] = 0;
			for(i=0; i<r+1; i++)

				row1[d] += 
				TAD(abs(((unsigned char)L->imageData[(y-r-1)*L->widthStep+3*i])-((unsigned char)R->imageData[(y-r-1)*R->widthStep+3*(i-min(i,d+offs_x))])) + 
				abs(((unsigned char)L->imageData[(y-r-1)*L->widthStep+3*i+1])-((unsigned char)R->imageData[(y-r-1)*R->widthStep+3*(i-min(i,d+offs_x))+1])) + 
				abs(((unsigned char)L->imageData[(y-r-1)*L->widthStep+3*i+2])-((unsigned char)R->imageData[(y-r-1)*R->widthStep+3*(i-min(i,d+offs_x))+2])),threshold);

			row2[d] = 0;
			for(i=0; i<r+1; i++)
				row2[d] += 
				TAD(abs(((unsigned char)L->imageData[(y+r)*L->widthStep+3*i])-((unsigned char)R->imageData[(y+r)*R->widthStep+3*(i-min(i,d+offs_x))])) + 
				abs(((unsigned char)L->imageData[(y+r)*L->widthStep+3*i+1])-((unsigned char)R->imageData[(y+r)*R->widthStep+3*(i-min(i,d+offs_x))+1])) + 
				abs(((unsigned char)L->imageData[(y+r)*L->widthStep+3*i+2])-((unsigned char)R->imageData[(y+r)*R->widthStep+3*(i-min(i,d+offs_x))+2])),threshold);

			acc[d] = pos0[d] - row1[d] + row2[d];
			pos0[d] = acc[d];

			out[y*w][d] = acc[d];
		}

		//FASE 1: bordo sinistro 
		for(x=1; x<r+1; x++){
			for(d=0; d<nDisp; d++){
				V[x+r][d] = V[x+r][d] 
				- (TAD(abs(((unsigned char)L->imageData[(y-r-1)*L->widthStep+3*(x+r)])-((unsigned char)R->imageData[(y-r-1)*R->widthStep+3*(x+r-min(x+r,d+offs_x))])) + 
				   abs(((unsigned char)L->imageData[(y-r-1)*L->widthStep+3*(x+r)+1])-((unsigned char)R->imageData[(y-r-1)*R->widthStep+3*(x+r-min(x+r,d+offs_x))+1])) + 
				   abs(((unsigned char)L->imageData[(y-r-1)*L->widthStep+3*(x+r)+2])-((unsigned char)R->imageData[(y-r-1)*R->widthStep+3*(x+r-min(x+r,d+offs_x))+2])),threshold)) 
					+ 
					TAD(abs(((unsigned char)L->imageData[(y+r)*L->widthStep+3*(x+r)])-((unsigned char)R->imageData[(y+r)*R->widthStep+3*(x+r-min(x+r,d+offs_x))])) + 
					abs(((unsigned char)L->imageData[(y+r)*L->widthStep+3*(x+r)+1])-((unsigned char)R->imageData[(y+r)*R->widthStep+3*(x+r-min(x+r,d+offs_x))+1])) + 
					abs(((unsigned char)L->imageData[(y+r)*L->widthStep+3*(x+r)+2])-((unsigned char)R->imageData[(y+r)*R->widthStep+3*(x+r-min(x+r,d+offs_x))+2])),threshold);
				acc[d] = acc[d] + V[x+r][d];
				out[y*w+x][d] = acc[d];
				V[x][d] = V[x][d] 
				- (TAD(abs(((unsigned char)L->imageData[(y-r-1)*L->widthStep+3*x])-((unsigned char)R->imageData[(y-r-1)*R->widthStep+3*(x-min(x,d+offs_x))])) + 
				   abs(((unsigned char)L->imageData[(y-r-1)*L->widthStep+3*x+1])-((unsigned char)R->imageData[(y-r-1)*R->widthStep+3*(x-min(x,d+offs_x))+1])) + 
				   abs(((unsigned char)L->imageData[(y-r-1)*L->widthStep+3*x+2])-((unsigned char)R->imageData[(y-r-1)*R->widthStep+3*(x-min(x,d+offs_x))+2])),threshold)) 
					+ 
					TAD(abs(((unsigned char)L->imageData[(y+r)*L->widthStep+3*x])-((unsigned char)R->imageData[(y+r)*R->widthStep+3*(x-min(x,d+offs_x))])) + 
					abs(((unsigned char)L->imageData[(y+r)*L->widthStep+3*x+1])-((unsigned char)R->imageData[(y+r)*R->widthStep+3*(x-min(x,d+offs_x))+1])) + 
					abs(((unsigned char)L->imageData[(y+r)*L->widthStep+3*x+2])-((unsigned char)R->imageData[(y+r)*R->widthStep+3*(x-min(x,d+offs_x))+2])),threshold);
			}
		}//x

		//FASE 2: parte centrale
		for(x=r+1; x<w-r; x++){
			for(d=0; d<nDisp; d++){
				V[x+r][d] = V[x+r][d] - 
					(TAD(abs(((unsigned char)L->imageData[(y-r-1)*L->widthStep+3*(x+r)])-((unsigned char)R->imageData[(y-r-1)*R->widthStep+3*(x+r-min(x+r,d+offs_x))])) + 
					 abs(((unsigned char)L->imageData[(y-r-1)*L->widthStep+3*(x+r)+1])-((unsigned char)R->imageData[(y-r-1)*R->widthStep+3*(x+r-min(x+r,d+offs_x))+1])) + 
					 abs(((unsigned char)L->imageData[(y-r-1)*L->widthStep+3*(x+r)+2])-((unsigned char)R->imageData[(y-r-1)*R->widthStep+3*(x+r-min(x+r,d+offs_x))+2])),threshold)) 
					+ 
					 TAD(abs(((unsigned char)L->imageData[(y+r)*L->widthStep+3*(x+r)])-((unsigned char)R->imageData[(y+r)*R->widthStep+3*(x+r-min(x+r,d+offs_x))])) + 
					 abs(((unsigned char)L->imageData[(y+r)*L->widthStep+3*(x+r)+1])-((unsigned char)R->imageData[(y+r)*R->widthStep+3*(x+r-min(x+r,d+offs_x))+1])) + 
					 abs(((unsigned char)L->imageData[(y+r)*L->widthStep+3*(x+r)+2])-((unsigned char)R->imageData[(y+r)*R->widthStep+3*(x+r-min(x+r,d+offs_x))+2])),threshold);

				acc[d] = acc[d] + V[x+r][d] - V[x-r-1][d];
				out[y*w+x][d] = acc[d];
			}
		}//x

		//FASE 3: bordo destro
		for(x=w-r; x<w; x++){
			for(d=0; d<nDisp; d++){
				acc[d] = acc[d] - V[x-r-1][d];
				out[y*w+x][d] = acc[d];
			}
		}//x
	}//y

	//######## BORDO IN BASSO ############
	for(y=h-r; y<h; y++){

		//FASE "0": PRIMA POSIZIONE A SX
		for(d=0; d<nDisp; d++){
			V[0][d] = V[0][d] - 
				(TAD(abs(((unsigned char)L->imageData[(y-r-1)*L->widthStep])-((unsigned char)R->imageData[(y-r-1)*R->widthStep])) + 
				 abs(((unsigned char)L->imageData[(y-r-1)*L->widthStep+1])-((unsigned char)R->imageData[(y-r-1)*R->widthStep+1])) + 
				 abs(((unsigned char)L->imageData[(y-r-1)*L->widthStep+2])-((unsigned char)R->imageData[(y-r-1)*R->widthStep+2])),threshold));

			row1[d] = 0;
			for(i=0; i<r+1; i++)
				row1[d] += 
				TAD(abs(((unsigned char)L->imageData[(y-r-1)*L->widthStep+3*i])-((unsigned char)R->imageData[(y-r-1)*R->widthStep+3*(i-min(i,d+offs_x))])) + 
				abs(((unsigned char)L->imageData[(y-r-1)*L->widthStep+3*i+1])-((unsigned char)R->imageData[(y-r-1)*R->widthStep+3*(i-min(i,d+offs_x))+1])) + 
				abs(((unsigned char)L->imageData[(y-r-1)*L->widthStep+3*i+2])-((unsigned char)R->imageData[(y-r-1)*R->widthStep+3*(i-min(i,d+offs_x))+2])),threshold);
			acc[d] = pos0[d] - row1[d];
			pos0[d] = acc[d];

			out[y*w][d] = acc[d];
		}

		//FASE 1: bordo sinistro
		for(x=1; x<r+1; x++){
			for(d=0; d<nDisp; d++){
				V[x+r][d] = V[x+r][d] - 
					(TAD(abs(((unsigned char)L->imageData[(y-r-1)*L->widthStep+3*(x+r)])-((unsigned char)R->imageData[(y-r-1)*R->widthStep+3*(x+r-min(x+r,d+offs_x))])) + 
					 abs(((unsigned char)L->imageData[(y-r-1)*L->widthStep+3*(x+r)+1])-((unsigned char)R->imageData[(y-r-1)*R->widthStep+3*(x+r-min(x+r,d+offs_x))+1])) + 
					 abs(((unsigned char)L->imageData[(y-r-1)*L->widthStep+3*(x+r)+2])-((unsigned char)R->imageData[(y-r-1)*R->widthStep+3*(x+r-min(x+r,d+offs_x))+2])),threshold));

				acc[d] = acc[d] + V[x+r][d];
				out[y*w+x][d] = acc[d];

				V[x][d] = V[x][d] - 
					(TAD(abs(((unsigned char)L->imageData[(y-r-1)*L->widthStep+3*x])-((unsigned char)R->imageData[(y-r-1)*R->widthStep+3*(x-min(x,d+offs_x))])) + 
					 abs(((unsigned char)L->imageData[(y-r-1)*L->widthStep+3*x+1])-((unsigned char)R->imageData[(y-r-1)*R->widthStep+3*(x-min(x,d+offs_x))+1])) + 
					 abs(((unsigned char)L->imageData[(y-r-1)*L->widthStep+3*x+2])-((unsigned char)R->imageData[(y-r-1)*R->widthStep+3*(x-min(x,d+offs_x))+2])),threshold));
			}
		}//x

		//FASE 2: parte centrale
		for(x=r+1; x<w-r; x++){
			for(d=0; d<nDisp; d++){
				V[x+r][d] = V[x+r][d] - 
					(TAD(abs(((unsigned char)L->imageData[(y-r-1)*L->widthStep+3*(x+r)])-((unsigned char)R->imageData[(y-r-1)*R->widthStep+3*(x+r-min(x+r,d+offs_x))])) + 
					 abs(((unsigned char)L->imageData[(y-r-1)*L->widthStep+3*(x+r)+1])-((unsigned char)R->imageData[(y-r-1)*R->widthStep+3*(x+r-min(x+r,d+offs_x))+1])) + 
					 abs(((unsigned char)L->imageData[(y-r-1)*L->widthStep+3*(x+r)+2])-((unsigned char)R->imageData[(y-r-1)*R->widthStep+3*(x+r-min(x+r,d+offs_x))+2])),threshold));
				acc[d] = acc[d] + V[x+r][d] - V[x-r-1][d];
				out[y*w+x][d] = acc[d];
			}
		}//x

		//FASE 3: bordo destro
		for(x=w-r; x<w; x++){
			for(d=0; d<nDisp; d++){
				acc[d] = acc[d] - V[x-r-1][d];
				out[y*w+x][d] = acc[d];
			}
		}//x
	}//y

	for(d=0; d<w;d++){
		free(V[d]);
	}
	free(V);
	free(acc);
	free(row1);
	free(row2);
	free(pos0);
}


// -----------------------------------------------------------------------------------------------------------------------
// -----------------------------------------------------------------------------------------------------------------------
// -----------------------------------------------------------------------------------------------------------------------


void box_filtering_sum_ACCV_2009(int w, int h, int ws, BYTE* raw, int Channel, int* out, int r){

int x,y,i,j;
int acc;

int row1, row2, pos0;

int *V;
V = (int *)calloc(w, sizeof(int));

int n = 2*r+1;
int sqn = n*n;

//############### PRIMA RIGA #################
//FASE "0": PRIMA POSIZIONE A SX
V[0]=0;
for(j=0; j<n; j++)
	V[0] += raw[j*ws+Channel];

acc=V[0];

for(i=1; i<n; i++)
for(j=0; j<n; j++)
	acc += raw[j*ws+3*i+Channel];
out[r*w+r] = acc;

pos0 = acc;

//FASE 1: calcolo colonne da aggiungere/togliere, aggiorno acc 
for(x=r+1; x<r+n; x++){
	V[x+r]=0;
	for(j=0; j<n; j++)
		V[x+r] += raw[j*ws+3*(x+r)+Channel];
	
	acc = acc + V[x+r] - V[x-r-1];
	out[r*w+x] = acc;

	V[x-r]=0;
	for(j=0; j<n; j++)
		V[x-r] += raw[j*ws+3*(x-r)+Channel];
}//x

//FASE 2: calcolo solo colonne da aggiungere (le altre le ho tutte)
for(x=r+n; x<w-r; x++){
	
	V[x+r]=0;
	for(j=0; j<n; j++)
		V[x+r] += raw[j*ws+3*(x+r)+Channel];
	
	acc = acc + V[x+r] - V[x-r-1];
	out[r*w+x] = acc;
}//x

//######## TUTTE LE RIGHE oltre alla prima ############

for(y=r+1; y<h-r; y++){

	//FASE "0": PRIMA POSIZIONE A SX
	//qui si pu ottimizzare il calcolo del primo acc rendendolo incrementale con posizione soprastante
	V[0] = V[0] - raw[(y-r-1)*ws+Channel] + raw[(y+r)*ws+Channel];

	row1 = 0;
	for(i=0; i<n; i++)
		row1 += raw[(y-r-1)*ws+3*i+Channel];
	row2 = 0;
	for(i=0; i<n; i++)
		row2 += raw[(y+r)*ws+3*i+Channel];
	acc = pos0 - row1 + row2;
	pos0 = acc;

	out[y*w+r] = acc;

	//FASE 1: calcolo colonne da aggiungere/togliere, aggiorno acc 
	for(x=r+1; x<r+n; x++){
		V[x+r] = V[x+r] - raw[(y-r-1)*ws+3*(x+r)+Channel] + raw[(y+r)*ws+3*(x+r)+Channel];
		acc = acc + V[x+r] - V[x-r-1];
		out[y*w+x] = acc;
		V[x-r] = V[x-r] - raw[(y-r-1)*ws+3*(x-r)+Channel] + raw[(y+r)*ws+3*(x-r)+Channel];
	}//x

	//FASE 2: calcolo solo colonne da aggiungere (le altre le ho tutte)
	for(x=r+n; x<w-r; x++){
		V[x+r] = V[x+r] - raw[(y-r-1)*ws+3*(x+r)+Channel] + raw[(y+r)*ws+3*(x+r)+Channel];
		acc = acc + V[x+r] - V[x-r-1];
		out[y*w+x] = acc;
	}//x
}//y

free(V);

}




// -----------------------------------------------------------------------------------------------------------------------
// -----------------------------------------------------------------------------------------------------------------------
// -----------------------------------------------------------------------------------------------------------------------



int FastStereoBilateral_ACCV_2009(IplImage *Left, IplImage *Right, IplImage *Result_LR, IplImage *Result_RL, IplImage *Result_LR_SMP, int dmin, int dmax, int radius, int radius_macro, float gamma_space, float gamma_color,int TAD_threshold, int scale)
{

	float number_of_macroblocks_check=((2.0*radius_macro+1)*(2.0*radius_macro+1))/((2.0*radius+1)*(2.0*radius+1));
	
	if (number_of_macroblocks_check != (int)number_of_macroblocks_check)
	{
		printf("Error: Something wrong with paprameters radius_macro and radius: aborting :-(...\n");
		return -1;
	}

	printf("Fast Bilateral Stereo is running...\n");

	int x,y, i_macro, j_macro, d;
	int Width, Height, WidthStep, WidthStepGray, border;

	int number_of_macroblocks, index_space_distance_weight;
	float *c1, *c2;


	float **color_distance_right;


	int **DSI, *R_Left, *R_Right, *G_Left, *G_Right, *B_Left, *B_Right;
	int R_right, G_right, B_right, R_left, G_left, B_left;;

	int min, disparity, area;
	float result, color_distance_left, c3, weight;
	double norm;

	// Init LUT
	init_exponential_lut_ACCV_2009(gamma_color, 255*255*3);	

	Width = Left->width;
	Height = Left->height;
	WidthStep = Left->widthStep;
	WidthStepGray = Result_LR->widthStep;
		
	border = 0;
	area = (2*radius+1)*(2*radius+1);
	number_of_macroblocks=((2*radius_macro+1)*(2*radius_macro+1))/((2*radius+1)*(2*radius+1));
	index_space_distance_weight=0;

		// Dynamic allocation

		c1 = (float*)calloc ((number_of_macroblocks), sizeof (float));
			if (c1==NULL) {printf("ABORTING... CAN'T ALLOCATE SPACE_DISTANCE_WEIGHT\n\n"); return -1;}

		DSI = (int**)calloc ((Width*Height), sizeof (int*));
			if (DSI==NULL) {printf("ABORTING... CAN'T ALLOCATE DSI MATRIX\n\n"); return -1;}
		for(x=0; x<Width*Height; x++)
			{DSI[x] = (int*)calloc(dmax+1, sizeof(int));
				if (DSI[x]==NULL) {printf("ABORTING... CAN'T ALLOCATE DSI VECTORS \n\n"); return -1;}}

		R_Left = (int *)calloc(Width*Height, sizeof(int));
			if(R_Left==NULL)	{printf("\nABORTING... CAN'T ALLOCATE R_LEFT\n\n"); return -1;}
		G_Left = (int *)calloc(Width*Height, sizeof(int));
			if(G_Left==NULL)	{printf("\nABORTING... CAN'T ALLOCATE G_LEFT\n\n"); return -1;}
		B_Left = (int *)calloc(Width*Height, sizeof(int));
			if(B_Left==NULL)	{printf("\nABORTING... CAN'T ALLOCATE B_LEFT\n\n"); return -1;}

		R_Right = (int *)calloc(Width*Height, sizeof(int));
			if(R_Right==NULL)	{printf("\nABORTING... CAN'T ALLOCATE R_RIGHT\n\n"); return -1;}
		G_Right = (int *)calloc(Width*Height, sizeof(int));
			if(G_Right==NULL)	{printf("\nABORTING... CAN'T ALLOCATE G_RIGHT\n\n"); return -1;}
		B_Right = (int *)calloc(Width*Height, sizeof(int));
			if(B_Right==NULL)	{printf("\nABORTING... CAN'T ALLOCATE B_RIGHT\n\n"); return -1;}

		c2 = (float*)calloc ((Height*number_of_macroblocks), sizeof (float));
			if (c2==NULL) {printf("ABORTING... CAN'T ALLOCATE SPACE_DISTANCE_REFERENCE\n\n"); return -1;}



		color_distance_right = (float **)calloc(Width, sizeof (float *));
			if(color_distance_right==NULL)	{printf("\nERRORE: IMPOSSIBILE ALLOCARE LA MATRICE COLOR_DISTANCE\n"); return -1;}

		for (x=0; x<Width; x++)		
			{color_distance_right[x] = (float *)calloc(number_of_macroblocks, sizeof(float));
				if(color_distance_right[x]==NULL)	{printf("\nERRORE: IMPOSSIBILE ALLOCARE I VETTORI DI COLOR_DISTANCE\n"); return -1;}}

		// Spatial weights computation

		for(j_macro=-radius_macro+radius; j_macro<=radius_macro-radius; j_macro=j_macro+(2*radius+1))
			for(i_macro=-radius_macro+radius; i_macro<=radius_macro-radius; i_macro=i_macro+(2*radius+1))
				{c1[index_space_distance_weight] = exp(-(sqrt((float((i_macro)*(i_macro) + (j_macro)*(j_macro))))/gamma_space));
				 index_space_distance_weight++;}



// TAD computation

box_filtering_stereo_color_TAD_single_channel_ACCV_2009(Width, Height, dmax+1, 0, Left, Right, DSI, radius, TAD_threshold);  


// Box filtering

{
box_filtering_sum_ACCV_2009(Width, Height, Left->widthStep,(unsigned char *)Left->imageData, 0,R_Left, radius);
box_filtering_sum_ACCV_2009(Width, Height, Left->widthStep,(unsigned char *)Left->imageData, 1,G_Left, radius);
box_filtering_sum_ACCV_2009(Width, Height, Left->widthStep,(unsigned char *)Left->imageData, 2,B_Left, radius);

box_filtering_sum_ACCV_2009(Width, Height, Right->widthStep,(unsigned char *)Right->imageData, 0,R_Right, radius);
box_filtering_sum_ACCV_2009(Width, Height, Right->widthStep,(unsigned char *)Right->imageData, 1,G_Right, radius);
box_filtering_sum_ACCV_2009(Width, Height, Right->widthStep,(unsigned char *)Right->imageData, 2,B_Right, radius);
}

// Box-filters normalization
for (x=0; x<Width*Height; x++)
   {R_Left[x] = R_Left[x]/area;
    G_Left[x] = G_Left[x]/area;
    B_Left[x] = B_Left[x]/area;
    R_Right[x] = R_Right[x]/area;
    G_Right[x] = G_Right[x]/area;
    B_Right[x] = B_Right[x]/area;
}


    bool updated=false;

memset(Result_LR->imageData, 255, sizeof(BYTE)*Result_LR->widthStep*Result_LR->height);
memset(Result_RL->imageData, 255, sizeof(BYTE)*Result_RL->widthStep*Result_RL->height);

for (y=0;y<Height;y++)
{
// ------------------ SMP ---------------------------------------------

#ifdef SMP_ON
	int *SMP_Collision_Dest = (int*)calloc (Width+(dmax-dmin+1), sizeof (int));	
	float *SMP_Collision_Dest_Score = (float*)calloc (Width+(dmax-dmin+1), sizeof (float));

	for (int q=0;q<Width+(dmax-dmin+1);q++) { 
		SMP_Collision_Dest_Score[q]=500000000;
		SMP_Collision_Dest[q]=-1; 
	}
#endif 

// ------------------ SMP ---------------------------------------------


	for (x=0; x<Width; x++)
	 {index_space_distance_weight=-1;
      for(j_macro=-radius_macro+radius; j_macro<=radius_macro-radius; j_macro=j_macro+(2*radius+1))
		 for(i_macro=-radius_macro+radius; i_macro<=radius_macro-radius; i_macro=i_macro+(2*radius+1))
		    {index_space_distance_weight++;

		  	 if (y+j_macro-radius<0) continue;
			 if (y+j_macro+radius>=Height) continue;
			 if (x+i_macro-radius-d<0) continue;
			 if (x+i_macro+radius>=Width) continue;

			 R_right=R_Right[(y+j_macro)*Width+i_macro+x]; 
			 G_right=G_Right[(y+j_macro)*Width+i_macro+x]; 
			 B_right=B_Right[(y+j_macro)*Width+i_macro+x]; 		

  			 color_distance_right[x][index_space_distance_weight] = 
				abs(R_right - ((unsigned char)Right->imageData[y*WidthStep+3*x])) * abs(R_right - ((unsigned char)Right->imageData[y*WidthStep+3*x])) +
				abs(G_right - ((unsigned char)Right->imageData[y*WidthStep+3*x+1])) * abs(G_right - ((unsigned char)Right->imageData[y*WidthStep+3*x+1])) +
				abs(B_right - ((unsigned char)Right->imageData[y*WidthStep+3*x+2])) * abs(B_right - ((unsigned char)Right->imageData[y*WidthStep+3*x+2]));}}

				

		     for (x=0; x<Width; x++)
			{
				if (y<border || y>Height-border || x<border || x>Width-border)
					{		
						continue;
					}

					
			min = 1000000;
		    disparity=-1;
			updated=false;

			// -------------- SUBPIXEL -----------------------------------
			double *score_vector = (double*)calloc ((dmax-dmin+1), sizeof (double));
			// -------------- SUBPIXEL -----------------------------------

				for (d=0; d<=dmax; d++)
				{
					score_vector[d]=-1; // subpixel init
					result=0.0;
				        norm=0.0;
					index_space_distance_weight=-1;

					for(j_macro=-radius_macro+radius; j_macro<=radius_macro-radius; j_macro=j_macro+(2*radius+1))
					for(i_macro=-radius_macro+radius; i_macro<=radius_macro-radius; i_macro=i_macro+(2*radius+1))
					{
						index_space_distance_weight++;

						if (y+j_macro-radius<0) continue;
						if (y+j_macro+radius>=Height) continue;
						if (x+i_macro-radius-d<0) continue;
						if (x+i_macro+radius>=Width) continue;
				 		if (x-d<0) continue;
					
						 if (d==dmin) 
						 {		
							// For reference image, computes weights once (i.e. d==0)
							R_left=R_Left[(y+j_macro)*Width+i_macro+x]; 
							G_left=G_Left[(y+j_macro)*Width+i_macro+x]; 
							B_left=B_Left[(y+j_macro)*Width+i_macro+x];  
		
							color_distance_left =  
							abs(R_left - ((unsigned char)Left->imageData[y*WidthStep+3*(x)])) * abs(R_left - ((unsigned char)Left->imageData[y*WidthStep+3*(x)])) +
							abs(G_left - ((unsigned char)Left->imageData[y*WidthStep+3*(x)+1])) * abs(G_left - ((unsigned char)Left->imageData[y*WidthStep+3*(x)+1])) +
							abs(B_left - ((unsigned char)Left->imageData[y*WidthStep+3*(x)+2]))*abs(B_left-((unsigned char)Left->imageData[y*WidthStep+3*(x)+2])) ;			
								 
							c2[y*number_of_macroblocks + index_space_distance_weight]=EXP_LUT_ACCV_2009[(int)color_distance_left];
						} // end dmin


				c3=EXP_LUT_ACCV_2009[(int)color_distance_right[x-d][index_space_distance_weight]];	

				weight=c1[index_space_distance_weight]*c2[y*number_of_macroblocks + index_space_distance_weight]*c3;

					norm += weight;
					result = result + DSI[(y+j_macro)*Width+x+i_macro][d]*weight;
					} // end loop_macro 					  
					
					result = result/norm;
					score_vector[d]=result;

					if (result<min) 
					{
						min = result;
					    	disparity=d;
						updated=true;
					}
				 } // end d

				 // ------------- subpixel -----------

				 float subpixel_adjustment=0;

				 if (updated==true)
				 {
					if (disparity>dmin && disparity<dmax && score_vector[disparity-1]>=0 && score_vector[disparity]>=0 && score_vector[disparity+1]>=0)
						subpixel_adjustment=get_subpixel_adjustment_min(score_vector[disparity-1], score_vector[disparity], score_vector[disparity+1]);
				 }
				 // ------------- subpixel -----------


				 if (updated==true) {
					 ((BYTE *)(Result_LR->imageData))[y*WidthStepGray+x]=(BYTE)((disparity+subpixel_adjustment)*scale);
					 ((BYTE *)(Result_RL->imageData))[y*WidthStepGray+x-disparity]=(BYTE)((disparity+subpixel_adjustment)*scale);
				 } 
				 else 
				 {
					 ((BYTE *)(Result_LR->imageData))[y*WidthStepGray+x]=255;
					 ((BYTE *)(Result_RL->imageData))[y*WidthStepGray+x-disparity]=255;
				 }

				 // ------------- subpixel -----------
				free(score_vector);
				 // ------------- subpixel -----------
	
// ----------- SMP -------------------------------

#ifdef SMP_ON

if (updated==true) {
	if (SMP_Collision_Dest[(x-disparity)]>0)
	 {
		// printf("SMP: collision (current) x =%d  d = %d score = %f VS point x=%d score = %f\n",x, disparity,min,SMP_Collision_Dest[(x-disparity)],SMP_Collision_Dest_Score[(x-disparity)]);
		if (SMP_Collision_Dest_Score[(x-disparity)]<min)  
			{	// previous was better...	
				// printf("Collision: previous was better ;-) ...\n");
				((BYTE *)(Result_LR_SMP->imageData))[y*WidthStepGray+x]=255;
			}
		else 
			{ // current is better...
				// printf("Collision: current is better->recovering ;-( ...\n");
				((BYTE *)(Result_LR_SMP->imageData))[y*WidthStepGray+x]=(disparity)*scale;
				((BYTE *)(Result_LR_SMP->imageData))[y*WidthStepGray-SMP_Collision_Dest[(x+disparity)]]=255;
			}
	 }
	else 
	{
		SMP_Collision_Dest[(x-disparity)]=disparity;
		SMP_Collision_Dest_Score[(x-disparity)]=min;
		((BYTE *)(Result_LR_SMP->imageData))[y*WidthStepGray+x]=(disparity)*scale;
	}

}

#endif 

// ----------- SMP -------------------------------

	} // end x

			 
	
	// ------ SMP -------------------

#ifdef SMP_ON

	free(SMP_Collision_Dest);
	free(SMP_Collision_Dest_Score);
	// ------ SMP -------------------

#endif 

	} // end y


// Deallocation

free(c1);

for (x=0; x<Width*Height;x++)		
free(DSI[x]);

free(DSI);

free(R_Left);
free(G_Left);
free(B_Left);

free(R_Right);
free(G_Right);
free(B_Right);

free(c2);



for (x=0; x<Width; x++)	 free(color_distance_right[x]);
free(color_distance_right);


return 0;

}




